const TreeNode = require('../TreeNode');
const Q_target_occupied = require('./Q_target_occupied');
const UseAbility = require('../UseAbility');
const BattleCalc = require('../../BattleCalc');
const AStarSearch = require('../../ai/AStarSearch/AStarSearch');
const { logError } = require('../../logger');

var attempts = 0;

const MoveDecision = function(actor, battleState) {
  const node = new TreeNode(actor, battleState);

  // branch indexes
  node.WITHIN_JUMP_RANGE = 0;
  node.NO_JUMP_ABILITY = 1;

  node.leap = BattleCalc.getEquippedAbility(actor, 'leap');
  node.teleport = BattleCalc.getEquippedAbility(actor, 'teleport');

  node.branches[node.WITHIN_JUMP_RANGE] = new Q_target_occupied(actor, battleState);
  node.branches[node.NO_JUMP_ABILITY] = new UseAbility(actor, 'move');

  node.aStarSearch = new AStarSearch();

  node.calculateMaxJumpRange = function() {
    // no need to compare ranges, because our design doesn't include any units with both leap and teleport
    if (this.leap)	{
      return this.leap.range.max;
    } else if (this.teleport) {
      return this.teleport.range.max;
    } else {
      return 0;
    }
  };

  node.targetBeyondJumpRange = function() {
    var leapAPCost = BattleCalc.getLeapAPCostByDist(this.actor, this.targetDist);
    var teleportMPCost = BattleCalc.getTeleportMPCostByDist(this.actor,this.targetDist);

    var leapTooFar = false;
    if (this.leap && (this.targetDist > this.leap.range.max || this.actor.ap < leapAPCost)) {
      leapTooFar = true;
    }

    var teleportTooFar = false;
    if (this.teleport && (this.targetDist > this.teleport.range.max || this.actor.mp < teleportMPCost)) {
      teleportTooFar = true;
    }

    var result = leapTooFar || teleportTooFar;
    return result;
  };

  node.getBestAdjacentTile = function(battleState) {
    let path = this.aStarSearch.getPath(
      battleState.getTileAtCoords(this.actor.x,this.actor.y),
      this.target,
      battleState,
    );

    if (path.length > 1) {
      return path[1]; // 0 being where the actor already is?
    } else {
      let tile;
      let adjacentTiles;
      let dist = BattleCalc.boardDistance(this.actor, this.target);

      var attempts = 0;
      try {
        while (
          (!path || path.length === 0) && dist > 0 && ++attempts < 30
        ) {
          adjacentTiles = battleState.getAdjacentTiles(this.target);
          for (let i = 0; i < adjacentTiles.length; ++i) {
            tile = adjacentTiles[i];
            if (BattleCalc.boardDistance(this.actor, tile) < dist) {
              this.target = tile;
              dist = BattleCalc.boardDistance(this.actor, this.target);
              path = this.aStarSearch.getPath(battleState.getTileAtCoords(this.actor.x,this.actor.y), this.target, battleState);
              break;
            }
          }
        }
      } catch (err) {
        logError(err, {
          module: 'MoveDecision',
          func: 'getBestAdjacentTile',
          actor,
          battleState,
        });
        return null;
      }

      if (path.length > 1) {
        return path[1];
      }

      return null;
    }
  };

  node.getNewTargetWithinRange = function() {
    const candidateTiles = [];

    for (var i in battleState.allTiles) {
      var tile = this.battleState.allTiles[ i ];
      var dist = BattleCalc.boardDistance(this.actor, tile);

      if (dist > 0 && dist <= this.calculateMaxJumpRange()
				&&	BattleCalc.pieceIsBetweenPair(tile,[this.actor,this.target]))
      {
        candidateTiles.push(tile);
      }
    }

    if (candidateTiles.length)	{
      return candidateTiles[ Math.floor(Math.random()*candidateTiles.length) ];
    } else {
      return null;
    }
  };

  node.query = function(input) {
    if (attempts > 10) {
      attempts = 0;
      return null;
    }

    if (!this.actor.inPlay
			||	input === this.actor
			||	input === this.battleState.getTileAtCoords(this.actor.x,this.actor.y)) {
      attempts = 0;
      return null;
    }

    // getRelevantTile must be defined by the "child" class
    this.target = this.getRelevantTile(input);
    if (!this.target) {
      attempts = 0;
      return null;
    }

    this.targetDist = BattleCalc.boardDistance(this.actor, this.target);

    this.jumpRange = this.calculateMaxJumpRange();

    if (this.targetDist === 0) {
      attempts = 0;
      return null;
    } else if (this.targetDist===1) {
      attempts = 0;
      return this.branches[this.NO_JUMP_ABILITY].query(this.target);
    } else if (this.jumpRange===0 || this.targetDist<3 || (this.actor.ai.priorities[0]==='support' && this.target.unit && this.target.team!==this.actor.team)) {
      var bestAdjacent = this.getBestAdjacentTile(this.battleState);
      if (bestAdjacent) {
        attempts = 0;
        return this.branches[this.NO_JUMP_ABILITY].query(bestAdjacent);
      } else {
        attempts = 0;
        return null;
      }

    } else if (this.targetBeyondJumpRange()) {
      ++attempts;
      return this.query(this.getNewTargetWithinRange());
    } else {
      attempts = 0;
      return this.branches[this.WITHIN_JUMP_RANGE].query(this.target);
    }
  };

  return node;
};
module.exports = MoveDecision;
